{

    TABLELOCK v1.2 1998
    By Germn Aracil ( human )

}

unit TableLock;

interface

uses
	Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
	DB, DBTables, BDE;

type
	TtlWaitMode = (mAllTime, mOnlyOne, mUser);

type
	TTableLock = class(TTable)
	private
		{ Private declarations }
// Bloqueo de registros
		FModeTableLock: TtlWaitMode;
		FLastRecordLocked: boolean;
		FPostToDisk: boolean;
// Propiedades aadidas para la gestin automtica de un registro nico auto contador verificado
		FContadorState :Boolean; // Activa y desactiva la gestin del contador
		FContadorUltimo :Variant;
		FContadorActual :Variant;
		FContadorTabla :TTable; // Tabla temporal
		FContadorField : String; // Nombre del registro a contar
	protected
		{ Protected declarations }
// Bloqueo de registros...
		function IsRecordLocked: boolean;
		procedure DoBeforeInsert; override;
		procedure DoBeforeEdit; override;
		procedure DoAfterPost; override;
// Para el contador
		procedure DoBeforeOpen; override;
		procedure DoBeforeClose; override;
		procedure DoBeforePost; override;
		function SrchContador: boolean;
	public
		{ Public declarations }
		constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;
	published
		{ Published declarations }
// Bloqueo de registros...
		property tlRecordIsLocked: boolean read IsRecordLocked;
		property tlRecordLockMode: TtlWaitmode
			read FModeTableLock write FModeTableLock;
		property tlLastRecordLocked: boolean read FLastRecordLocked;
		property tlPostToDisk: boolean
			read FPostToDisk write FPostToDisk;
// Propiedad aadida para la gestion de un campo autocontado y nico.
		property cntState :Boolean read FContadorState write FContadorState;
		property cntFieldName :String read FContadorField write FContadorField;
	end;

  procedure Register;

implementation

uses
	MessageRecordLocked;

procedure Register;
begin
	RegisterComponents('HDBPlus', [TTableLock]);
end;

constructor TTableLock.Create(AOwner: TComponent);
begin
	Inherited;
	FContadorTabla := TTable.Create(nil);
end;

destructor TTableLock.Destroy;
begin
	inherited Destroy;
end;

// Examina el record actual de la tabla para ver si est bloqueado
function TTableLock.IsRecordLocked: boolean;
var
	Locked: BOOL;
	hCur: hDBICur;
	rslt: DBIResult;

begin
	UpdateCursorPos;
	// Si est bloqueado por la sesin actual...
	Check(DbiIsRecordLocked(Handle, Locked));
	Result := Locked;
	// If the current session does not have a lock and the ByAnyone varable is
	// set to check all sessions, continue check...
	if (Result = False) then
	begin
		// Get a new cursor to the same record...
		Check(DbiCloneCursor(Handle, False, False, hCur));
		try
			// Try and get the record with a write lock...
			rslt := DbiGetRecord(hCur, dbiWRITELOCK, nil, nil);
			if rslt <> DBIERR_NONE then
			begin
				// if an error occured and it is a lock error, return true...
				if HiByte(rslt) = ERRCAT_LOCKCONFLICT then
					Result := True
				else
					// If some other error happened, throw an exception...
					Check(rslt);
			end
			else
				// Release the lock in this session if the function was successful...
				Check(DbiRelRecordLock(hCur, False));
		finally
			// Close the cloned cursor...
			Check(DbiCloseCursor(hCur));
		end;
	end;
end;

// Verifica el dato del campo a contar
function TTableLock.SrchContador: boolean;
begin
	with FContadorTabla do begin
		if (RecordCount = 0) then
    	if self.FieldByName(FContadorField).Value = null then begin
				FContadorUltimo := 1;
	      Result := True;
				exit;
			end else begin
      	Result := False; // Tiene algo asignado
        Exit;
      end;

		setkey;
		FContadorTabla.FieldByName(FContadorField).Value := Self.FieldByName(FContadorField).Value;
		Result := GotoKey;
		if (self.FieldByName(FContadorField).value = null) or
			(self.FieldByName(FContadorField).value = 0) then begin
			Result := True;
		end;
	end;
end;

// Copia la tabla en la tabla del contador
procedure TTableLock.DoBeforeOpen;
begin
	inherited;
	FContadorTabla.DataBaseName := DataBaseName;
	FContadorTabla.SessionName := SessionName;
	FContadorTabla.TableName := TableName;
	FContadorTabla.IndexFieldNames := IndexFieldNames;
	FContadorTabla.IndexName := IndexName;
	FContadorTabla.Open;
end;

procedure TTableLock.DoBeforeClose;
begin
	inherited;
	FContadorTabla.Close;
end;

// Recoge el valor mximo del field a contar (Tiene que estr indexada por el)
procedure TTAbleLock.DobeforeInsert;
begin
// Si est activado
	inherited;
end;

// Actualiza la tabla dependiendo del modo elegido
// mAllTime: se espera hasta que se desbloque el registro
// mOnlyOne: Sale sin actualizar la tabla
// mUser: muestra un cuadro de dialogo para que el usuario decida que hacer
// Al salir LastRecordLocked indica con True si se encontro bloqueo.
procedure TTableLock.DoBeforeEdit;
var
	Second : Double;
begin
	inherited;
// Recogemos el valor del registro autocontado si esta activada la opcin
	if FContadorState then begin
		FContadorTabla.Last;
		FContadorActual := FieldByName(FContadorField).Value;
	end;
// Primero hay que ver si el record est bloqueado
	FLastRecordLocked := IsRecordLocked;
	if not FLastRecordLocked then exit;
	if FModeTableLock = mOnlyOne then exit;
	if FModeTableLock = mAllTime then begin
		FormMessageRecordLocked := TFormMessageRecordLocked.Create ( self );
		FormMessageRecordLocked.Refresh;
		FormMessageRecordLocked.SetFocus;
		while IsRecordLocked do begin
			Second := Now;
			repeat until (Second + 1e-4) < Now; // Pausa
		end;
		FormMessageRecordLocked.Free;
	end;
	if FModeTableLock = mUser then
		repeat // Decide el usuario
			if Application.MessageBox('El registro no se puede actualizar por estar'+
			' bloqueado por otro usuario.'+#13+' Desea intentarlo otra vez ?',
			'ATENCIN', mb_YesNo) = IDNo then begin
				Cancel;
				break;
			end;
		until not IsRecordLocked;
	Refresh;
end;

// Comprueba antes de grabar los datos que el contador (si es activo) es correcto
procedure TTableLock.DoBeforePost;
begin
// Si el contador es activado comprueva su valor
	if FContadorState then begin
// Dependiendo de que est en modo edicin o insercin
		if State = dsEdit then begin
// Llamamos a SrchContador para ver si existe en la tabla
			if SrchContador then
				FieldByName(FContadorField).Value := FContadorActual;
// en caso de que est en modo insert
		end else begin
    	with TQuery.Create(self) do begin
        DataBaseName := Self.DatabaseName;
        SessionName := Self.SessionName;
        SQL.Clear;
        SQL.Add('SELECT MAX('+FContadorField+') FROM '+Self.TableName);
        Open;
				FContadorUltimo := Fields[0].AsVariant + 1;
      	Free;
      end;
			if SrchContador then
									FieldByNAme(FContadorField).Value := FContadorUltimo;
    end;
	end;
	inherited DoBeforePost ;
end;


// Graba cambios al disco si FPostToDisk = true
procedure TTableLock.DoAfterPost;
begin

	inherited DoAfterPost;
	if not FPostToDisk then exit;
	DbiSaveChanges(Handle);

end;
end.
